# pylint: disable=missing-module-docstring
from maps4fs.generator.dtm.arctic import ArcticProvider
from maps4fs.generator.dtm.baden import BadenWurttembergProvider
from maps4fs.generator.dtm.bavaria import BavariaProvider
from maps4fs.generator.dtm.canada import CanadaProvider
from maps4fs.generator.dtm.czech import CzechProvider
from maps4fs.generator.dtm.denmark import DenmarkProvider
from maps4fs.generator.dtm.england import England1MProvider
from maps4fs.generator.dtm.finland import FinlandProvider
from maps4fs.generator.dtm.flanders import FlandersProvider
from maps4fs.generator.dtm.france import FranceProvider
from maps4fs.generator.dtm.hessen import HessenProvider
from maps4fs.generator.dtm.italy import ItalyProvider
from maps4fs.generator.dtm.lithuania import LithuaniaProvider
from maps4fs.generator.dtm.mv import MecklenburgVorpommernProvider
from maps4fs.generator.dtm.niedersachsen import NiedersachsenProvider
from maps4fs.generator.dtm.norway import NorwayProvider
from maps4fs.generator.dtm.nrw import NRWProvider
from maps4fs.generator.dtm.rema import REMAProvider
from maps4fs.generator.dtm.sachsenanhalt import SachsenAnhaltProvider
from maps4fs.generator.dtm.scotland import ScotlandProvider
from maps4fs.generator.dtm.spain import SpainProvider
from maps4fs.generator.dtm.srtm import SRTM30Provider
from maps4fs.generator.dtm.switzerland import SwitzerlandProvider
from maps4fs.generator.dtm.thuringia import ThuringiaProvider
from maps4fs.generator.dtm.usgs_wcs import USGSWCSProvider
